/*
	Mac_Pong II.d from 
	Whitesmiths C on 	DEC PDP-11 to 
	Megamax C on 		Mac 128K to 
	Lightspeed C on 	Mac SE to 
	Think C 5.0 on 		Mac IIcx. 
	
	Copyright 1992 David L. O'Connor 
	07/08/93 Think C 6.0
*/

#include "Sound.h"
#include "animation.h"

/* Ball speeds */
#define SPEED_1		1
#define SPEED_2		3
#define SPEED_3		5
#define SPEED_4 		7
#define SPEED_5		9
#define SPEED_6		11
#define SPEED_7		13
#define SPEED_8		15
#define SPEED_9		17

/* the PICT resource id #'s */
#define screen_pict	1
#define ball_pict	2
#define lpad_pict	3
#define rpad_pict	4

/* the menu resource id #'s */
#define	appleid 128	
#define	fileid  129
#define	editid  130
#define	skillid 131
#define	soundid 132
#define	speedid 133

/* the window resource id #'s */
#define	game_window 128	

#define HIGHSCORE			21
#define GAME_IN_PROGRESS	(l_paddle.score > 0 || r_paddle.score > 0)
#define TRAVELING_LEFT          (ball.direction >= 90 && ball.direction <= 270)
#define TRAVELING_RIGHT        ((ball.direction >=  0 && ball.direction <= 90) || (ball.direction >= 270 && ball.direction <= 360))
#define TRACK_ERROR	        46
#define SPEED_FACTOR            (ball.speed / 3)

typedef struct {
    int count;
    int amplitude;
    int duration;
} tone;
typedef tone tones[5001];

typedef struct {
	spriterecord sprite;
	int score;
	} paddle;
	
typedef struct {
	int mode;
	tone triplet[1];
} bleep_tag;
		
typedef struct {
	int mode;
	tone triplet[2];
} blat_tag;
	
bleep_tag bleep_buf;
blat_tag  blat_buf;
paddle l_paddle, r_paddle;
spriterecord ball;
WindowPtr gamewindow, which_window;
WindowRecord winstorage;
Rect r, dragrect;
Point pad_loc, screen_center;
EventRecord gameevent;
MenuHandle gamemenu[7];
Str255 menutitle;
int skill_level, done, paused, last_won, volleys, sound_on,
	top_wall, bottom_wall, chosen_speed, cursor_visible, pad_low_limit;
unsigned char *title = "\p00                         Mac_Pong IId                         00";

main(void) 
{
	setup();
	while (! done)
		handle_events();
	FlushEvents(everyEvent, 0);
	StopSound();
	cleanupsprites();
}

setup(void)
{
	static unsigned long secs;
	
	init_mac();
	initsounds();
	SetRect(&dragrect, qd.screenBits.bounds.left + 4, qd.screenBits.bounds.top + 4, qd.screenBits.bounds.right - 4, qd.screenBits.bounds.bottom - 4);
	gamewindow = GetNewWindow ( game_window, &winstorage, (WindowPtr) -1L );
	GetDateTime( &secs);
	randSeed = secs;
	init_game();
	build_menus();
	create_screen();
	setupsprites();
	create_paddles();
	restart_game();
}

/* pretty much straight from SAMP in I.M. */
handle_events(void)
{
	if (GetNextEvent(everyEvent, &gameevent)){
		switch (gameevent.what){
			case mouseDown :
				switch (FindWindow(gameevent.where, &which_window)){
					case inMenuBar :
						docommand(MenuSelect(gameevent.where));
					break;
					case inGoAway:
						if (TrackGoAway(gamewindow, gameevent.where))
							done = 1;
					break;
					case inSysWindow :
						SystemClick(&gameevent, which_window);
					break;
					case inDrag :
						DragWindow(which_window, gameevent.where, &dragrect);
					break;
					case inContent :
						if (which_window != FrontWindow())
							SelectWindow(which_window);
					break;
				}
			break;
			case activateEvt :
				if (gameevent.modifiers & activeFlag)
					disable_edit_menu();
				else
					enable_edit_menu();
			break;
			case updateEvt :
				SetPort(gamewindow);
				BeginUpdate(gamewindow);
				display(&gamewindow->portRect, &gamewindow->portBits);
				EndUpdate(gamewindow);
			break;
		}
	}
	SystemTask();
	check_cursor();	
	play_pong();
}

disable_edit_menu(void)
{
	DisableItem(gamemenu[2], 1);
	DisableItem(gamemenu[2], 3);
	DisableItem(gamemenu[2], 4);
	DisableItem(gamemenu[2], 5);
	DisableItem(gamemenu[2], 6);
}

enable_edit_menu(void)
{
	EnableItem(gamemenu[2], 1);
	EnableItem(gamemenu[2], 3);
	EnableItem(gamemenu[2], 4);
	EnableItem(gamemenu[2], 5);
	EnableItem(gamemenu[2], 6);
}

/* hide cursor inside game window's visable region */
check_cursor(void)
{
	GlobalToLocal(&gameevent.where);
	if (PtInRgn(gameevent.where, gamewindow->visRgn)){
		if (cursor_visible){
			cursor_visible = False;
			HideCursor();
		}
	}
	else{
		if (! cursor_visible){
			cursor_visible = True;
			ShowCursor();
		}
	}
}

play_pong(void)
{	
	if ( (! paused) && (l_paddle.score < HIGHSCORE && r_paddle.score < HIGHSCORE)){
		if (ball.status){
			check_status();
			animatesprite(&ball, next_loc(&ball), &gamewindow->portBits);
		}
		else
			serve_ball();
		Move_left_paddle();
		Move_right_paddle();
	}
}

initsounds(void){
	bleep_buf.mode = blat_buf.mode = swMode;
	bleep_buf.triplet[0].count = 1000;
	bleep_buf.triplet[0].amplitude = 255;
	bleep_buf.triplet[0].duration = 5;
	blat_buf.triplet[0].count = 1000;
	blat_buf.triplet[0].amplitude = 255;
	blat_buf.triplet[0].duration = 5;
	blat_buf.triplet[1].count = 3000;
	blat_buf.triplet[1].amplitude = 255;
	blat_buf.triplet[1].duration = 10; 
}

init_game(void)
{
	ball.status        = Dead;
	done                 = 0;
	skill_level        = 3;
	sound_on          = 0;
	chosen_speed   = SPEED_5;
	last_won          = RIGHT;
	cursor_visible  = True;
}

build_menus(void)
{	
	register int i;
	
	InitMenus();
	gamemenu[0] = GetMenu(appleid);
	AddResMenu(gamemenu[0],'DRVR');
	gamemenu[1] = GetMenu(fileid);
	gamemenu[2] = GetMenu(editid);
	gamemenu[3] = GetMenu(skillid);
	gamemenu[4] = GetMenu(soundid);
	gamemenu[5] = GetMenu(speedid);
	for(i = 0; i < 6; i++)
		InsertMenu(gamemenu[i], 0);
	disable_edit_menu();
	CheckItem(gamemenu[3], skill_level, 1);
	CheckItem(gamemenu[5], 5, 1);
	DrawMenuBar();
	SetPort(gamewindow);
}

docommand(menu_selection)
long menu_selection;
{
	register int the_Item, old_speed, i;
	static Str255 name;
	
	the_Item = LoWord(menu_selection);
	switch (HiWord(menu_selection)){
		case appleid :
			if (the_Item == 1){
				ShowCursor();
				Alert(128, 0L);
			}
			else{
				GetItem(gamemenu[0], the_Item, name);
				OpenDeskAcc(name);
				SetPort(gamewindow);
			}
		break;
		case editid:
			SystemEdit(the_Item - 1);
		break;
		case fileid:
			switch (the_Item){
				case 1 :
					if (paused){
						paused = 0;
						SetItem(gamemenu[1], 1, "\pPause");
					}
					else{
						paused = 1;
						SetItem(gamemenu[1], 1, "\pContinue");
					}
				break;
				case 2:
					restart_game();
				break;
				case 3 :
					done = 1;
				break;
			}
		break;
		case skillid :
			CheckItem(gamemenu[3], skill_level, 0);
			skill_level = the_Item;
			CheckItem(gamemenu[3], skill_level, 1);
		break;
		case soundid:
			if (sound_on){
				sound_on = 0;
				SetItem(gamemenu[4], 1, "\pTurn On");
					}
			else{
				sound_on = 1;
				SetItem(gamemenu[4], 1, "\pTurn Off");
			}
		break;
		case speedid:
			for (i = 1; i < 10; i++)
				CheckItem(gamemenu[5], i, 0);
			old_speed = chosen_speed;
			CheckItem(gamemenu[5], the_Item, 1);
			switch (the_Item){
				case 1 :
					chosen_speed = SPEED_1;
				break;
				case 2:
					chosen_speed = SPEED_2;
				break;
				case 3 :
					chosen_speed = SPEED_3;
				break;
				case 4 :
					chosen_speed = SPEED_4;
				break;
				case 5 :
					chosen_speed = SPEED_5;
				break;
				case 6:
					chosen_speed = SPEED_6;
				break;
				case 7 :
					chosen_speed = SPEED_7;
				break;
				case 8 :
					chosen_speed = SPEED_8;
				break;
				case 9 :
					chosen_speed = SPEED_9;
				break;
			}
			ball.speed -= (old_speed - chosen_speed);
		break;
	}
	HiliteMenu(0);
}

create_paddles(void)
{	
	pad_low_limit = gamewindow->portRect.bottom - l_paddle.sprite.pic_v;
	pad_loc.v = screen_center.v - (l_paddle.sprite.pic_v / 2);
	pad_loc.h = gamewindow->portRect.left + 1;
	animatesprite(&l_paddle.sprite, pad_loc, &gamewindow->portBits);
	pad_loc.v = screen_center.v - (r_paddle.sprite.pic_v / 2);
	pad_loc.h = gamewindow->portRect.right - 1 - r_paddle.sprite.pic_h;
	animatesprite(&r_paddle.sprite, pad_loc, &gamewindow->portBits);
}	

serve_ball(void)
{
	register Point serve_point;
	static duration = 0;
	
	duration++;
	if (duration > 50 ){
		duration = 0;
		if (! createsprite(&ball, ball_pict))
			badpic();
		randomize_right_paddle();
		serve_point.v = screen_center.v - (ball.pic_v / 2);
		serve_point.h = screen_center.h - (ball.pic_h / 2);
		animatesprite(&ball, serve_point, &gamewindow->portBits);
		if (GAME_IN_PROGRESS)
			ball.direction = (last_won == RIGHT) ? 180 : 0;
		else
			ball.direction = (Random() > 0) ? 180 : 0;
		ball.speed  = chosen_speed;
		ball.status = Alive;
		bleep();
	}
}

randomize_right_paddle(void)
{
	pad_loc = topLeft(r_paddle.sprite.drag_r);
	pad_loc.v = screen_center.v - (r_paddle.sprite.pic_v / 2);
	r_paddle.sprite.speed = Random() > 0 ? 2  : 4;
	if (Random() > 0 )
		pad_loc.v += r_paddle.sprite.speed;
	else
		pad_loc.v -= r_paddle.sprite.speed;		 
	keep_paddle_in_bounds(&pad_loc);
	animatesprite(&r_paddle.sprite, pad_loc, &gamewindow->portBits);
	pad_loc = topLeft(r_paddle.sprite.drag_r);
	if ((pad_loc.v < screen_center.v) &&  (pad_loc.v + r_paddle.sprite.pic_v > screen_center.v)){
		if (Random() > 0 )
			pad_loc.v += r_paddle.sprite.pic_v;
		else
			pad_loc.v -= r_paddle.sprite.pic_v;
		animatesprite(&r_paddle.sprite, pad_loc, &gamewindow->portBits);
	}
}

/* someone scored a point */
kill_ball(void)
{
	killsprite(&ball, True);
	display(&ball.update_r, &gamewindow->portBits);
	Move_right_paddle();
	Move_left_paddle();
	ball.status = Dead;
	volleys = 0;
	if (GAME_IN_PROGRESS)
		blat();
	display_score();
}

restart_game(void)
{
	l_paddle.score = r_paddle.score = 0;
	kill_ball();
}

/* check for bounces, direction changes, scoring, etc */
check_status(void)
{	

	/* COMBINED_HEIGHT = l_paddle.sprite.pic_v + ball.pic_v 
	   in this implementation it is 56, so define it for efficiency */
	#define COMBINED_HEIGHT 56
	
	register ball_top	= ball.drag_r.top;
	register ball_bottom = ball.drag_r.bottom;
	register ball_left	= ball.drag_r.left;
	register ball_right	= ball.drag_r.right;
	
	/* make it a little harder as time goes by */
	ball.speed = chosen_speed;
	if (volleys > 25)
		ball.speed += 10;
	else if (volleys > 20)
		ball.speed += 8;
	else if (volleys > 15)
		ball.speed += 6;
	else if (volleys > 10)
		ball.speed += 4;
	else if (volleys > 5)
		ball.speed += 2;
		
	/* the ball and the left boundry */
	if (ball_left < l_paddle.sprite.drag_r.right && TRAVELING_LEFT){
		if (ball_bottom >= l_paddle.sprite.drag_r.top && ball_top <= l_paddle.sprite.drag_r.bottom){
			volleys++;
			bleep();
			ball.direction = COMBINED_HEIGHT - ((ball_bottom - l_paddle.sprite.drag_r.top) * 2);
			if (ball.direction < 0)
				ball.direction += 360;
		}
		else{
			last_won = RIGHT;
			r_paddle.score++;
			kill_ball();
		}
		return;
	}
	
	/* the ball and the right boundry. */
	if (ball_right > r_paddle.sprite.drag_r.left && TRAVELING_RIGHT){
		if (ball_bottom >= r_paddle.sprite.drag_r.top && ball_top <= r_paddle.sprite.drag_r.bottom){
			volleys++;
			bleep();
			ball.direction = 180 - COMBINED_HEIGHT + ((ball_bottom - r_paddle.sprite.drag_r.top) * 2);
		}
		else{
			last_won = LEFT;
			l_paddle.score++;
			kill_ball();
		}
		return;
	}

	/* the ball and the top wall */
	if ( (ball_top <= top_wall) && (ball.direction > 0 && ball.direction < 180) ){
			ball.direction = 360 - ball.direction;
		bleep();
		return;
	}
	
	/* the ball and the bottom wall */
	if ( (ball_bottom >= bottom_wall) && (ball.direction > 180 && ball.direction < 360)){
		ball.direction = 360 - ball.direction;
		bleep();
		return;
	}
}

Move_left_paddle(void)
{	
	static Point mouse_loc;
	
	pad_loc = topLeft(l_paddle.sprite.drag_r);
	GetMouse(&mouse_loc);
	pad_loc.v = mouse_loc.v;
	keep_paddle_in_bounds(&pad_loc);
	animatesprite(&l_paddle.sprite, pad_loc, &gamewindow->portBits);
}

Move_right_paddle(void)
{
	pad_loc = topLeft(r_paddle.sprite.drag_r);
	r_paddle.sprite.speed = ball.speed + 1;
	if ( (ball.drag_r.right > screen_center.h) && TRAVELING_RIGHT){
		if (ball.drag_r.top + handicap() < r_paddle.sprite.drag_r.top)
			pad_loc.v -= r_paddle.sprite.speed;
		else if (ball.drag_r.bottom - handicap() > r_paddle.sprite.drag_r.bottom)
			pad_loc.v += r_paddle.sprite.speed;
	}
	keep_paddle_in_bounds(&pad_loc);
	animatesprite(&r_paddle.sprite, pad_loc, &gamewindow->portBits);
}

display_score(void)
{
	convert_score_to_string((title + 1),  l_paddle.score);
	convert_score_to_string((title + 65), r_paddle.score);
	SetWTitle(gamewindow, title);
}

convert_score_to_string(s, i)
register unsigned char *s;
register int i;
{
	static Str255 temp;
	
	NumToString((long)i, temp);
	PtoCstr(temp);
	if (i <= 9){
		temp[1] = temp[0];
		temp[0] = '0';
	}
	*s		 = temp[0];
	*(s + 1) = temp[1];
}	

keep_paddle_in_bounds(pad_loc)
register Point *pad_loc;
{
	if (pad_loc->v < 0)
		pad_loc->v = 0;
	else if (pad_loc->v > pad_low_limit)
		pad_loc->v = pad_low_limit;
}

bleep(void)
{
	if (sound_on){
		if (! SoundDone())
			StopSound();
		StartSound((Ptr) &bleep_buf, (long) sizeof(bleep_buf),   0L);
	}
}

blat(void)
{
	if (sound_on){
		if (! SoundDone())
			StopSound();
		StartSound((Ptr) &blat_buf, (long) sizeof(blat_buf),   0L);
	}
}

/* Every so often, let the Mac's paddle fail to track the ball until
   the ball has passed it by a certain amount, depending on ball's speed.
   This is the heart of a satisfying game. */
int handicap(void)
{
	register mac_skill = skill_level + 1;
	
	return (Random() % mac_skill) == 0 ? TRACK_ERROR + SPEED_FACTOR: 0;
}

/************************* Sprite Stuff ******************************/

setupsprites(void)
{	
	if(initsprites(&gamewindow->portRect) == False)
		badinit();
	if (! createsprite(&l_paddle.sprite, lpad_pict))
		badpic();
	if (! createsprite(&r_paddle.sprite, rpad_pict))
		badpic();
}

create_screen(void)
{	
	register PicHandle screen_image;
	
	top_wall 	    = gamewindow->portRect.top    + 25;
	bottom_wall     = gamewindow->portRect.bottom - 25;
	screen_center.h = (gamewindow->portRect.right  - gamewindow->portRect.left) / 2;
	screen_center.v = (gamewindow->portRect.bottom - gamewindow->portRect.top)  / 2;
	screen_image    = (PicHandle) GetPicture(screen_pict);
	if (screen_image == Nil)
		badpic();
	GetNextEvent(everyEvent, &gameevent); /* for Multifinder */
	GetNextEvent(everyEvent, &gameevent); 
	GetNextEvent(everyEvent, &gameevent);
	GetNextEvent(everyEvent, &gameevent);
	DrawPicture(screen_image, &((**screen_image).picFrame));
	ReleaseResource((Handle) screen_image);
}

badinit(void)
{		
	ShowCursor();
	Alert(129, Nil);
	ExitToShell();
}

badpic(void)
{		
	ShowCursor();
	Alert(130, Nil);
	ExitToShell();
}

cleanupsprites(void)
{
	killsprite(&ball, True);
	killsprite(&l_paddle.sprite, True);
	killsprite(&r_paddle.sprite, True);
	closesprites(True);
}
