/*
	Animation functions inspired by MacTutor July 1987 article 
	in Think C. 
	Copyright 1993 David L. O'Connor
	04/16/93 Think C 5.0
*/

#include "animation.h"

/******************************************************************************/

 static double	sin_table[361];
 static double	cos_table[361];
 static GrafPtr	oldport, junkport;
 BitMap		offscreenbits;
 
/******************************************************************************/

init_mac()
{
	MaxApplZone();					/* This call expands our heap to its maximum value. */
	MoreMasters();				/* Allocate additional Master Handles */
	InitGraf(&qd.thePort); 			/* init quickdraw */
	InitFonts();					/* init font manager */
	InitWindows();					/* init window manager */
	InitMenus();					/* init menu manager */
	TEInit();						/* init TextEdit */
	InitDialogs(0L);					/* init dialog manager */
	FlushEvents(everyEvent,0);		/* get rid of left over events */
	InitCursor();					/* set cursor to arrow */
}

/******************************************************************************/

/* Create a new bitmap for a given rectangle's size.  Return
   the address of the bitmap if successful, Nil if not.
 */
Ptr newbitmap(bitmap, rect)
register BitMap	*bitmap;
register Rect	*rect;
{
	static Size mapsize;
	
	bitmap->rowBytes = ((rect->right - rect->left + 31) / 32) * 4;
	mapsize = ((Size) bitmap->rowBytes * ((Size) rect->bottom - (Size) rect->top));
	bitmap->baseAddr = NewPtr(mapsize);
	(*bitmap).bounds = *rect;
	return (MemError() == noErr) ? bitmap->baseAddr : Nil;
}

/******************************************************************************/

/* Set up our environment, a junk Grafport and Offscreen Bitmap. 
	Mac Plus ROM (version # 117)  or greater required to support 
   'CalcMask' and 'CopyMask'. If compiled for 68881, check if one is present
 */
initsprites(screen_rect)
register Rect *screen_rect;
{		
	
	static short rom, machine;
	
#if __option(mc68881)
	SysEnvRec	this_mac;
	SysEnvirons ( 1, &this_mac);
	if (! this_mac.hasFPU)
		return False;
#endif
	
	Environs(&rom, &machine);
	if (rom < 117)
		return False;
		
	GetPort(&oldport);
	if (newbitmap(&offscreenbits, screen_rect) == Nil)
		return False;
	junkport = (GrafPtr) NewPtr(sizeof (GrafPort));
	if (junkport == Nil){
		DisposPtr(offscreenbits.baseAddr);
		return False;
	}
	OpenPort(junkport);
	RectRgn((*thePort).visRgn, &(**((*thePort).clipRgn)).rgnBBox);
	CopyBits(&oldport->portBits, &offscreenbits, &oldport->portRect,
			 &oldport->portRect, srcCopy, Nil);
	SetPort(oldport);
	init_tables();
	return True;
}

/******************************************************************************/

/* create a new movable object from a PICT resource ID #. NOTE: PICT
   resources must have an rectangle origion of 0,0 for proper operation.
 */
createsprite(sprite, picid, speed, direction)
register spriterecord *sprite;		/* the movable object		*/
register int picid;				/* the picture resource id		*/
register int speed, direction;
{
	register PicHandle pic;
	
	GetPort(&oldport);
	SetPort(junkport);
		
/*	Get the picture using the PICT resource id#.
 */
	pic = (PicHandle) GetPicture(picid);
	if (pic == Nil)
		return False;
	MoveHHi((Handle) pic);
	HLock((Handle) pic);
	sprite->status = Alive;
	
/* Allocate the bitmaps to the size of the picture's boundry.
 */
	if ( (newbitmap(&sprite->picturebits, &(**pic).picFrame) == Nil) ||
	     (newbitmap(&sprite->picturemask, &(**pic).picFrame) == Nil) ||
	     (newbitmap(&sprite->underbits,   &(**pic).picFrame) == Nil) ){
			HUnlock((Handle) pic);
			ReleaseResource((Handle) pic);
			killsprite(sprite, False);
			return False;
	}

/* Save the picture's height and width and midpoints for future use.
 */
	sprite->pic_h = sprite->picturebits.bounds.right  - sprite->picturebits.bounds.left;
	sprite->pic_v = sprite->picturebits.bounds.bottom - sprite->picturebits.bounds.top;
			
/* Draw the picture in the picturebits.
 */
	SetPortBits(&sprite->picturebits);
	EraseRect(&sprite->picturebits.bounds);
	DrawPicture(pic, &((**pic).picFrame));

/* Create the picture's mask. Note: If boundry of picture has openings,
   the image will appear ghostly.
 */
	CalcMask(sprite->picturebits.baseAddr, sprite->picturemask.baseAddr, sprite->picturebits.rowBytes, 
			sprite->picturemask.rowBytes, sprite->pic_v, sprite->picturebits.rowBytes / 2);	
			 		 
#ifdef USE_COPYBITS
/* Create the mask's and underbits regions for CopyBits. Move underbits region to proper position.
*/
	sprite->maskRgn = NewRgn();
	BitMapToRegion(sprite->maskRgn,&sprite->picturemask);
	sprite->underbitsRgn = NewRgn();
	BitMapToRegion(sprite->underbitsRgn,&sprite->picturemask);
	OffsetRgn(sprite->underbitsRgn, sprite->underbits.bounds.left - (**sprite->underbitsRgn).rgnBBox.left,
 						  	  sprite->underbits.bounds.top - (**sprite->underbitsRgn).rgnBBox.top);
#endif

/* initialize speed and direction
 */
 	sprite->speed 	  = speed;
 	sprite->direction = direction;
 
/* Finish up.
 */
	HUnlock((Handle) pic);
	ReleaseResource((Handle) pic);
	SetPort(oldport);
	return True;
}

/******************************************************************************/

/* Set a sprite's direction
*/
set_direction(sprite, angle)
register spriterecord *sprite;
register angle;
{
	sprite->direction = angle;
}

/******************************************************************************/

/* Increment a sprite's angle of direction
*/
increment_direction(sprite, amount)
register spriterecord *sprite;
register amount;
{
	sprite->direction += amount;
	if (sprite->direction > 360)
		sprite->direction -= 360;
}

/******************************************************************************/

/* Decrement a sprite's angle of direction
*/
decrement_direction(sprite, amount)
register spriterecord *sprite;
register amount;
{
	sprite->direction -= amount;
		if (sprite->direction < 0)
		sprite->direction += 360;
}

/******************************************************************************/

/* Set a sprite's speed
 */
set_speed(sprite, speed)
register spriterecord *sprite;
register speed;
{
	sprite->speed = speed;
}

/******************************************************************************/

/* Initialize the sine and cosine tables. Note the reversal of sine values to 
	match Quickdraw's inverted Y axis scale. A direction of 90 will now move 
	an object vertically up the screen. Put up a modal dialog info box, as this can be a long process without an FPU.
*/
init_tables()
{
	register DialogPtr theDialog;
	register angle, inverse;
 	register double one_eighty = 180.0;
  	register double  Pi     	 = 3.14159265358979323846;
	
	theDialog = GetNewDialog ( 128, (Ptr) 0L, (WindowPtr) -1L );
	DrawDialog ( theDialog );
	for (angle = 0, inverse = 360; angle < 361; angle++, inverse--){
		sin_table[angle] = sin(((double) inverse / one_eighty) * Pi);
		cos_table[angle] = cos(((double) angle   / one_eighty) * Pi);
		}
	DisposDialog ( theDialog );
}

/******************************************************************************/

/* Return sprite's next location based on it's speed and direction.
 */
Point next_loc(sprite)
register spriterecord *sprite;
{
	register Point  sprite_loc;
	register double speed = sprite->speed;
	
	sprite_loc.h = sprite->drag_r.left + round(cos_table [sprite->direction] * speed);
	sprite_loc.v = sprite->drag_r.top  + round(sin_table [sprite->direction] * speed);
	return sprite_loc;
}

/******************************************************************************/

int round(double_value)
register double double_value;
{
	register int truncated_value = double_value;
	register double rounding       = double_value - truncated_value;
	
	if( rounding >= 0.5)
		truncated_value++;
	else if (rounding <= -0.5 )
		truncated_value--;
	 return truncated_value;
}

/******************************************************************************/

/* Copy a rectangle from the offscreen bitmap to the screen */
display (r_ptr, dest_bit_map )
register Rect   *r_ptr;
register BitMap	*dest_bit_map;
{
	CopyBits(&offscreenbits, dest_bit_map, r_ptr, r_ptr, srcCopy, Nil);
}

/******************************************************************************/

/* Move a sprite to it's new location.
 */
movesprite(sprite, location)
register spriterecord *sprite;
register Point location;
{	
 	if (sprite->status == Dead)
		return;
 
/* Move our object's rectangle .
 */
	sprite->drag_r.top	     = sprite->drag_r.bottom = location.v;
	sprite->drag_r.left	     = sprite->drag_r.right  = location.h;
	sprite->drag_r.bottom += sprite->pic_v; 
	sprite->drag_r.right    += sprite->pic_h;
	
/* Create the update rectangle as the union of where we were,
   and where we've gone. This will be used by our display function.
 */
	UnionRect(&sprite->drag_r, &sprite->last_r, &sprite->update_r);

/* Save stuff where our picture will be offscreen, and draw the picture.
*/

#ifdef USE_COPYBITS
	CopyBits(&offscreenbits,&sprite->underbits,&sprite->drag_r,&sprite->underbits.bounds,
			srcCopy,sprite->underbitsRgn);
	OffsetRgn(sprite->maskRgn, sprite->drag_r.left - (**sprite->maskRgn).rgnBBox.left,
						    sprite->drag_r.top - (**sprite->maskRgn).rgnBBox.top);
	CopyBits(&sprite->picturebits,&offscreenbits,&sprite->picturebits.bounds,&sprite->drag_r,
			srcCopy,sprite->maskRgn);
#else
	CopyMask(&offscreenbits, &sprite->picturemask, &sprite->underbits, 
			 &sprite->drag_r, &sprite->picturemask.bounds, &sprite->underbits.bounds);
	CopyMask(&sprite->picturebits, &sprite->picturemask, &offscreenbits,
			 &sprite->picturebits.bounds, &sprite->picturemask.bounds, &sprite->drag_r);
#endif
}

/******************************************************************************/

/* After the screen display, fix things up offscreen for the next time.
 */
fixupsprite(sprite)
register spriterecord *sprite;
{
	if (sprite->status == Dead)
		return;
		
 /* Remember what will need fixing next time around.
 */
 	sprite->last_r = sprite->drag_r;
 	
/* FIx up where our picture was offscreen.
*/
#ifdef USE_COPYBITS
	CopyBits(&sprite->underbits,&offscreenbits,&sprite->underbits.bounds,&sprite->drag_r,
			 srcCopy,sprite->maskRgn);
#else 
	CopyMask(&sprite->underbits, &sprite->picturemask, &offscreenbits,
			 &sprite->underbits.bounds, &sprite->picturemask.bounds, &sprite->drag_r);
#endif
}

/******************************************************************************/

/* Move, display and fixup.
 */
animatesprite(sprite, location, dest_bit_map)
register spriterecord *sprite;
register Point location;
register BitMap	*dest_bit_map;
{			 
	if (sprite->status == Dead)
		return;
	
/* Move our object's rectangle.
 */
	sprite->drag_r.top = sprite->drag_r.bottom = location.v;
	sprite->drag_r.left = sprite->drag_r.right = location.h;
	sprite->drag_r.bottom += sprite->pic_v; 
	sprite->drag_r.right    += sprite->pic_h;
	
/* Create the update rectangle as the union of where we were,
   and where we've gone. This will be used by our display function.
 */
	UnionRect(&sprite->drag_r, &sprite->last_r, &sprite->update_r);

/* Save stuff where our picture will be offscreen, and draw the picture.
*/		 
 #ifdef USE_COPYBITS
	CopyBits(&offscreenbits,&sprite->underbits,&sprite->drag_r,&sprite->underbits.bounds,
			srcCopy,sprite->underbitsRgn);
	OffsetRgn(sprite->maskRgn, sprite->drag_r.left - (**sprite->maskRgn).rgnBBox.left,
						    sprite->drag_r.top - (**sprite->maskRgn).rgnBBox.top);
	CopyBits(&sprite->picturebits,&offscreenbits,&sprite->picturebits.bounds,&sprite->drag_r,
			 srcCopy,sprite->maskRgn);
#else
	CopyMask(&offscreenbits, &sprite->picturemask, &sprite->underbits, 
			 &sprite->drag_r, &sprite->picturemask.bounds, &sprite->underbits.bounds);
	CopyMask(&sprite->picturebits, &sprite->picturemask, &offscreenbits,
			 &sprite->picturebits.bounds, &sprite->picturemask.bounds, &sprite->drag_r);
#endif
	
/* Display it on the screen.
 */
	CopyBits(&offscreenbits, dest_bit_map, &sprite->update_r, 
			   &sprite->update_r, srcCopy, Nil);
			
/* Remember what will need fixing next time around.
 */
 	sprite->last_r = sprite->drag_r;
 	
/* FIx up where our picture was offscreen.
*/
#ifdef USE_COPYBITS
	CopyBits(&sprite->underbits,&offscreenbits,&sprite->underbits.bounds,&sprite->drag_r,
			 srcCopy,sprite->maskRgn);
#else 
	CopyMask(&sprite->underbits, &sprite->picturemask, &offscreenbits,
			 &sprite->underbits.bounds, &sprite->picturemask.bounds, &sprite->drag_r);
#endif
}

/******************************************************************************/

/* Deallocate everything associated with a sprite, optionally remove his
   picture from the offscreen bitmap.
 */
killsprite(sprite, removepic)
register spriterecord *sprite;
register Boolean removepic;
{
	if (sprite->status == Dead)
		return;
	if (removepic)
		fixupsprite(sprite);
	if (sprite->picturebits.baseAddr != Nil)
		DisposPtr(sprite->picturebits.baseAddr);
	if (sprite->picturemask.baseAddr != Nil)
		DisposPtr(sprite->picturemask.baseAddr);
	if (sprite->underbits.baseAddr != Nil)
		DisposPtr(sprite->underbits.baseAddr);
# ifdef USE_COPYBITS
	DisposeRgn(sprite->underbitsRgn);
	DisposeRgn(sprite->maskRgn);
#endif
	sprite->status = Dead;
}

/******************************************************************************/

/* Deallocate our environment.
 */
closesprites(disposebitmap)
register Boolean disposebitmap;
{	
	if (junkport != Nil){
		ClosePort(junkport);
		DisposPtr((Ptr) junkport);
	}
	if (disposebitmap)
		if (offscreenbits.baseAddr != Nil)
			DisposPtr(offscreenbits.baseAddr);
}

/******************************************************************************/
